
open Puzzle
open Board

module Cells = struct
  module Iset = Set.Make(struct type t = int
				let compare = Pervasives.compare end)
  type t = Iset.t
  let empty = Iset.empty
  let is_empty = Iset.is_empty
  let add i j t = Iset.add ((i lsl 4) lor j) t
  let mem i j t = Iset.mem ((i lsl 4) lor j) t
  let remove i j t = Iset.remove ((i lsl 4) lor j) t
  let iter f t = Iset.iter (fun n -> f (n lsr 4) (n land 0b1111)) t

end

let choose_mincard b todo =
  let best_c = ref max_int in
  let best = ref (-1,0,Pset.empty) in
  let test i j =
    let ps = Board.poss b i j in
    let c = Pset.cardinal ps in
    if c < !best_c then begin
      best_c := c;
      best := (i,j,ps)
    end
  in
  Cells.iter test todo;
  match !best with
    | (-1,_,_) -> assert false
    | s -> s

let iter f ~draw ~clear ~choose b todo =
  let rec loop b todo = 
    if Cells.is_empty todo then 
      f b
    else 
      let i,j,ps = choose b todo in
      let todo = Cells.remove i j todo in
      Pset.random_iter
	(fun p -> 
	   try 
	     let b = place ~complete:true b i j p in 
	     draw i j p;
	     try loop b todo with Incorrect | Incomplete -> clear i j
	   with Incorrect | Incomplete -> 
	     ())
	ps
  in
  Cells.iter (fun i j -> assert (not (is_real_piece (piece_at b i j)))) todo;
  loop b todo

let enum b todo =
  let count = ref 0 in
  iter 
    (fun _ -> incr count) (fun _ _ _ -> ()) (fun _ _ -> ())
    choose_mincard b todo;
  !count

