
open Puzzle

(*
module Pset = struct
  include 
    Set.Make
    (struct 
       type t = piece
       let compare p1 p2 = 
	 (*Pervasives.compare (p1.id,p1.rot) (p2.id,p2.rot)*)
	 Pervasives.compare 
	   ((p1.id lsl 2) lor p1.rot) ((p2.id lsl 2) lor p2.rot) 
     end)

  let of_array a =
    Array.fold_left (fun s p -> add p s) empty a
end
*)

type t = { 
  board : piece Pmatrix.t;
  todo  : int; (* number of empty cells *)
  score : int;
  poss  : Pset.t Pmatrix.t; (* possibilities *)
  await : int Parray.t;
  avail : int Parray.t;
}

let initial_poss =
  Pmatrix.init n n
    (fun i j -> 
       if i = 0 && j = 0 then
	 Pset.of_array corners_rot1
       else if i = 0 && j = n-1 then
	 Pset.of_array corners_rot2
       else if i = n-1 && j = 0 then
	 Pset.of_array corners_rot0
       else if i = n-1 && j = n-1 then
	 Pset.of_array corners_rot3
       else if i = 0 then
	 Pset.of_array edges_rot2
       else if i = n-1 then
	 Pset.of_array edges_rot0
       else if j = 0 then
	 Pset.of_array edges_rot1
       else if j = n-1 then
	 Pset.of_array edges_rot3
       else
	 Pset.union 
	   (Pset.union 
	      (Pset.of_array others_rot0)
	      (Pset.of_array others_rot1))
	   (Pset.union 
	      (Pset.of_array others_rot2)
	      (Pset.of_array others_rot3))
    )

let initial_avail =
  let avail = Array.create 23 0 in
  let add p = avail.(p) <- avail.(p) + 1 in
  Array.iter (fun p -> add p.right; add p.up; add p.left; add p.down) pieces;
  Parray.init 23 (fun i -> avail.(i))

let empty = {
  board = Pmatrix.init n n (fun _ _ -> dummy_piece);
  todo = n * n;
  score = 0;
  poss = initial_poss;
  await = Parray.create 23 0;
  avail = initial_avail;
}

let score b = b.score

let todo b = b.todo

let avail b = b.avail

let piece_at b i j = Pmatrix.get b.board i j

let poss b i j = Pmatrix.get b.poss i j

exception Incorrect
exception Incomplete

(* applies filter [f] on [poss[i,j]] *)
let filter ~complete i j f poss =
  let ps = Pmatrix.get poss i j in
  match Pset.filter_opt f ps with
    | None -> 
	poss
    | Some ps ->
	if complete && Pset.is_empty ps then raise Incomplete;
	Pmatrix.set poss i j ps

let filter_p ~complete i0 j0 p board poss0 = 
  Pmatrix.map
    (fun i j ps ->
       if is_real_piece (Pmatrix.get board i j) then
	 ps
       else begin
	 match Pset.filter_opt (fun q -> q.id != p.id) ps with
	   | None -> 
	       ps
	   | Some ps ->
	       if (i != i0 || j != j0) && Pset.is_empty ps && complete then 
		 raise Incomplete;
	       ps
       end)
    poss0

let place ~complete b i j p =
  let poss = filter_p ~complete i j p b.board b.poss in
  let score = ref b.score in
  let piece = Pmatrix.get b.board in
  let poss = 
    if j > 0 then begin
      let l = piece i (j-1) in
      if not (is_real_piece l) then
	filter ~complete i (j-1) (fun q -> q.right == p.left) poss
      else
	if l.right == p.left then (incr score; poss) else raise Incorrect;
    end else
      poss
  in
  let poss = 
    if j < n-1 then begin
      let r = piece i (j+1) in
      if not (is_real_piece r) then
	filter ~complete i (j+1) (fun q -> q.left == p.right) poss
      else
	if r.left == p.right then (incr score; poss) else raise Incorrect;
    end else
      poss
  in
  let poss =
    if i > 0 then begin
      let d = piece (i-1) j in
      if not (is_real_piece d) then
	filter ~complete (i-1) j (fun q -> q.up == p.down) poss
      else
	if d.up == p.down then (incr score; poss) else raise Incorrect;
    end else
      poss
  in
  let poss = 
    if i < n-1 then begin
      let u = piece (i+1) j in
      if not (is_real_piece u) then
	filter ~complete (i+1) j (fun q -> q.down == p.up) poss
      else
	if u.down == p.up then (incr score; poss) else raise Incorrect;
    end else
      poss
  in
  let add_free await avail p =
    let await_p = Parray.get await p + 1 in
    let avail_p = Parray.get avail p - 1 in
    if complete && avail_p < await_p then raise Incomplete;
    Parray.set await p await_p, Parray.set avail p avail_p
  in
  let add_bound await avail p =
    let await_p = Parray.get await p - 1 in
    let avail_p = Parray.get avail p - 1 in
    if complete && avail_p < await_p then raise Incomplete;
    Parray.set await p await_p, Parray.set avail p avail_p
  in
  let await,avail =
    if complete then
      let await,avail =
	if i > 0 then
	  if not (is_real_piece (piece (i-1) j)) then 
	    add_free b.await b.avail p.down 
	  else
	    add_bound b.await b.avail p.down
	else
	  b.await,b.avail
      in
      let await,avail =
	if i < n-1 then
	  if not (is_real_piece (piece (i+1) j)) then 
	    add_free await avail p.up 
	  else
	    add_bound await avail p.up
	else
	  await,avail
      in
      let await,avail =
	if j > 0 then
	  if not (is_real_piece (piece i (j-1))) then 
	    add_free await avail p.left 
	  else
	    add_bound await avail p.left
	else
	  await,avail
      in
      if j < n-1 then
	if not (is_real_piece (piece i (j+1))) then 
	  add_free await avail p.right 
	else
	  add_bound await avail p.right
      else
	await,avail
    else
      b.await,b.avail
  in
(**
  if !score > !best then begin
    printf "score = %d@." !score;
    best := !score;
    if !score > 200 then dump !score
  end;
**)
  { board = Pmatrix.set b.board i j p;
    score = !score;
    todo = b.todo - 1;
    poss = poss;
    await = await;
    avail = avail; 
  }


let initial = 
  if is_eternity_ii then 
    let b = place ~complete:true empty 7 7 others_rot2.(139-1-4-56) in
    place ~complete:true b 2 2 others_rot1.(181-1-4-56) 
  else 
    empty

let create ~complete sol =
  let b = ref empty in
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      if sol.(i).(j) != dummy_piece then 
	b := place ~complete !b i j sol.(i).(j)
    done 
  done;
  !b

let iter_empty f b =
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      (*if not (is_real_piece (Pmatrix.get b.board i j)) then f i j*)
      if Pmatrix.get b.board i j == dummy_piece then f i j
    done 
  done

let save b file =
  let sol = Array.create_matrix n n (-1,0) in
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      let p = piece_at b i j in
      sol.(i).(j) <- (p.id,p.rot)
    done 
  done;
  Solution.save n sol file
