open Format
open Brute7

let iteratively_expand_tree first_tile first_rot b =
  let time = ref (Unix.time ()) in
  try
    while true do
      b := expand_tree_root init_board first_tile first_rot !b;
      let n = size !b in
      report_iteration !b;
      let t = Unix.time () in
      if t -. !time > 3600.0 (* 60 minutes *) then
	begin
	  let f = "dump_" ^ (string_of_int first_tile) ^ "_" 
	    ^ (string_of_int first_rot) ^ "_" ^(string_of_int n) in
	  dump_tree (Filename.concat sol_dir f) !b;
	  printf "saved dump file %s@." f;
	  time := t;
	end;
    done
  with
      Exit ->
	assert false

module Game = struct

  module HC = Hashtbl.Make(Network.Client)
  let client_table = HC.create 7

  let new_client c = HC.add client_table c (ref Idle); true

  type message = computing_server_message

  let receive c m = 
    let answer msg =
      Network.send c.Network.Client.fd (Msg msg)
    in
    let state = HC.find client_table c in
    match m,!state with
      | Compute (t,x,y), Idle -> 
	  printf "Start computation for tile=%d, rot=%d@." x y;
	  let t = ref t in
	  let pid = 
	    Thread.create (iteratively_expand_tree x y) t
	  in
	  state := Computing(pid,t,x,y)
      | Compute _,Computing _ ->
	  answer "Cannot start a new computation will computing"
      | Stop,Computing(pid,t,x,y) -> 
	  Thread.kill pid;
	  state := Idle;
	  answer "Computation stopped"
      | Stop, Idle ->
	  answer "Cannot stop: already idle"
      | Stats, Idle -> 
	  answer "Cannot give stats: i am idle"
      | Stats, Computing(pid,t,x,y) -> 
	  let t = !t in
	  Network.send c.Network.Client.fd (Status(x,y,score t,size t))

  let each_tick () = ()

  let send c = () 
(*
    let s = HC.find client_table c in
    Network.send c.Network.Client.fd !s
*)

  let disconnection c = ()

end

module Server = Network.Make(Game)

(* main launch *)

let () =
  begin
    printf "Starting server@.";
    Server.run ~port:2007 ~ack:false ~timeout:(-1.0)
  end


