
open Format
open Graphics
open Puzzle
open Board

let graphics = ref false
let greedy = ref false
let size = ref 2
let load = ref None
let sol_dir = ref "best2"
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-load", Arg.String (fun f -> load := Some f), "<file>  load state";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-greedy", Arg.Set greedy, "search for incomplete solution";
   "-size", Arg.Set_int size, "edge size";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let () = Arg.parse spec (fun _ -> ()) 
  "brute.opt [-p puzzle] [-load state] [-graphics] [-greedy]"
let graphics = !graphics
let greedy = !greedy
let load = !load
let sol_dir = !sol_dir
let size = !size

module MinCard = struct
  (* choose a piece with minimal possibilities cardinality *)
  let choose_in f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  let choose b =
    choose_in (fun i j -> i < size || i > 15-size || j < size || j > 15-size) b
 
end

let step = 50
let half = step / 2
let scale = half-1

let () =
  if graphics then Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let display b =
  for i = 0 to n-1 do
    for j = 0 to n-1 do
      if i < size || i > 15-size || j < size || j > 15-size then
	draw_tile i j (piece_at b i j)
    done
  done;
  ignore (read_key ())

let count = ref 0
let nb = (16-2*size)*(16-2*size)-1 (*16 * 4 + 12 * 4*)

let rec loop b = 
  if todo b = nb then begin 
    incr count; 
    if graphics then display b;
    if !count mod 10 = 0 then printf "%d @?" !count;
    raise Incomplete 
  end;
  let i,j,ps = MinCard.choose b in
  Pset.iter
    (fun p -> 
       try 
	 let b = place ~complete:true b i j p in loop b 
       with Incorrect | Incomplete -> 
	 ())
    ps;
  raise Incomplete
  
let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let init_board = match load with
  | None -> 
      Board.initial
  | Some f ->       
      let lsol = Puzzle.load f in
      let b = Board.create ~complete:true lsol in
      b

let () = 
  begin try loop init_board with Incomplete -> () end;
  printf "%d solutions@." !count
