
(* loads a solution, displays it and computes the score *)

open Format
open Graphics
open Puzzle

let file = ref None
let () = 
  Arg.parse ["-p", Arg.String (fun _ -> ()), ""]
    (fun f -> file := Some f) "check [-p file.data] file.sol"
let file = match !file with
  | None -> eprintf "check  [-p file.data] file.sol"; exit 1 
  | Some f -> f

let step = 56
let half = step / 2
let scale = half+1

let () = Draw.open_board ~title:"Check" ~lines:false step

let vertical_line i j =
  set_color white;
  set_line_width 3;
  moveto (step+(j+1)*step) (step+i*step); 
  lineto (step+(j+1)*step) (step+(i+1)*step)
  
let hor_line i j =
  set_color white;
  set_line_width 3;
  moveto (step+j*step) (step+(i+1)*step); 
  lineto (step+(j+1)*step) (step+(i+1)*step)
 
let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let sol = load file 

(* graphical display *)

let () =
  for i = 0 to n-1 do
    for j = 0 to n-1 do
      if sol.(i).(j) != dummy_piece then draw_tile i j sol.(i).(j) else clear_tile i j
    done
  done

(* score *)

let score = ref 0

let () =
  for i = 0 to n-1 do
    for j = 0 to n-2 do
      if sol.(i).(j) != dummy_piece && sol.(i).(j+1) != dummy_piece &&
	 sol.(i).(j).right == sol.(i).(j+1).left 
      then incr score
      else vertical_line i j
    done;
    if i < n-1 then
      for j = 0 to n-1 do
	if sol.(i).(j) != dummy_piece && sol.(i+1).(j) != dummy_piece &&
	  sol.(i).(j).up == sol.(i+1).(j).down 
	then incr score
	else hor_line i j
      done
  done;
  printf "score = %d@." !score;
  set_window_title ("Check: file " ^ file ^ ", score: " ^ string_of_int !score)


let () = ignore (read_key ())
