
open Graphics
open Format
open Puzzle
open Board
open Backt

let step = 50
let half = step / 2
let scale = half-1

let () = Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let () = if is_eternity_ii then draw_tile 7 7 (piece_at initial 7 7)

let todo =
  let s = ref Cells.empty in
  for i = 0 to 15 do
    for j = 0 to 15 do
      if i = j || i = j+1 || i = j-1 then s := Cells.add i j !s
    done
  done;
  Cells.remove 7 7 !s

let check b =
  for i = 0 to 15 do
    for j = 0 to 15 do
      let p = piece_at b i j in
      if is_real_piece p then begin
	if j > 1 then begin
	  let l = piece_at b i (j-1) in
	  if is_real_piece l then assert (l.right = p.left)
	end;
	if j < 15 then begin
	  let r = piece_at b i (j+1) in
	  if is_real_piece r then assert (r.left = p.right)
	end;
	if i > 1 then begin
	  let d = piece_at b (i-1) j in
	  if is_real_piece d then assert (d.up = p.down)
	end;
	if i < 15 then begin
	  let u = piece_at b (i+1) j in
	  if is_real_piece u then assert (u.down = p.up)
	end;
      end
    done
  done

let display b =
  for i = 0 to 15 do
    for j = 0 to 15 do
      let p = piece_at b i j in
      if is_real_piece p then draw_tile i j p else clear_tile i j
    done
  done;
  ignore (read_key ())

let choose b todo =
  (* display b; *)
  let i,j,ps = choose_mincard b todo in
  let ps =
    Pset.filter 
      (fun p -> 
	(i = 0 && j = 0) ||
	  (i = 1 && j = 0) ||
	  (i = 15 && j = 15) ||
	  (i = 15 && j = 14) ||
	  ((i = j || i = j+1) && p.down = p.right) ||
	  (i = j-1))
      ps
  in
  i,j,ps

let () =
  iter
    (fun b -> printf ".@?")
    (*draw_tile clear_tile *)
    (fun _ _ _ -> ()) (fun _ _ -> ())
    choose 
    Board.initial
    todo    
