
open Format
open Graphics
open Puzzle0
open Board0

let load = ref None
let sol_dir = ref "edit3"
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-d", Arg.Set_string sol_dir, "<dir>  directory for solutions (default 'edit3')";]
let usage =   "editor3.opt [-p puzzle] [-d dir] file"
let () = Arg.parse spec (fun f -> load := Some f) usage
let load = match !load with
  | None -> Arg.usage spec usage; exit 1
  | Some f -> f

let sol_dir = !sol_dir


let step = 56
let half = step / 2
let scale = ref (half+1)

let () = 
  open_graph (sprintf " %dx%d" ((n+6) * step) ((n+2) * step))


let vertical_line i j =
  set_color white;
  set_line_width 3;
  moveto (step+(j+1)*step) (step+i*step); 
  lineto (step+(j+1)*step) (step+(i+1)*step)
  
let hor_line i j =
  set_color white;
  set_line_width 3;
  moveto (step+j*step) (step+(i+1)*step); 
  lineto (step+(j+1)*step) (step+(i+1)*step)
 
let draw_tile i j t r =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.right x y !scale (Puzzle0.right_pat t r);
  Draw.up x y !scale (Puzzle0.up_pat t r);
  Draw.left x y !scale (Puzzle0.left_pat t r);
  Draw.down x y !scale (Puzzle0.down_pat t r)


let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect x y step step

let border_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_line_width 1;
  set_color cyan;
  draw_rect (x+1) (y+1) (step-2) (step-2)

(*

let () = 
  if graphics then begin
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	draw_tile i j pieces.(n*i+j)
      done
    done;
    ignore (read_key ());
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	clear_tile i j
      done
    done;
    if is_eternity_ii then draw_tile 7 7 (piece_at initial 7 7)
  end

let best = ref 0

module Spiral = struct
    
  let i = ref 7
  let j = ref 7
    
  let pos = Array.create 255 (0,0)
    
  let store =
    let k = ref (-1) in
    fun () -> 
      incr k;
      pos.(!k) <- !i,!j
	
  let up () = incr i
  let down () = decr i
  let right () = incr j
  let left () = decr j
  let repeat n f = for i = 1 to n do f (); store () done
    
  let () =
    for k = 1 to 15 do
      if k mod 2 = 1 then begin
	repeat k up; repeat k right;
      end else begin 
	repeat k down; repeat k left
      end
    done;
    repeat 15 down

  let choose1 b = 
    let rec loop k =
      let i,j = pos.(k) in 
      if piece_at b i j == dummy_piece then i,j else loop (k+1)
    in
    let i,j = loop 0 in i, j, poss b i j
      
  let choose2 b =
    let rec loop k =
      let i,j = pos.(k) in 
      if piece_at b i j == dummy_piece then i,j else loop (k-1)
    in
    let i,j = loop 254 in i, j, poss b i j

  let alternate b =
    if (todo b) mod 2 = 0 then choose1 b else choose2 b 

end

module MinCard = struct
  (* choose a piece with minimal possibilities cardinality *)
  let choose_in f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal possibilities cardinality *)
  let choose b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      let ps = poss b i j in
      let c = Pset.cardinal ps in
      if c < !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s

  let choose2 b =
    if todo b > 230 then Spiral.choose1 b else choose b

  (* choose a piece with minimal possibilities cardinality
     apart from the 3x3 region around (7,7) *)
  let choose_apart77 = 
    choose_in (fun i j -> i < 6 || i > 8 || j < 6 || j > 8)

  let choose3 b =
    if todo b < 20 then choose b else choose_apart77 b

  let avoid = [3,3; 3,11; 11,3; 11,11; 8,8]

  let choose4 b =
    (*try choose_in (fun i j -> j mod 8 = 7) b*)
    (*try choose_in (fun i j -> j <= 8 && i <= 8) b*)
    (*try choose_in (fun i j -> abs (i-7)<=4 && abs(j-7)<=2) b*)
    (*try 
      if (todo b) mod 10 <= 5 then 
	choose_in (fun i j -> i+j <= 12) b
      else
	choose_in (fun i j -> i+j >= 14) b*)
    (*try choose_in (fun i j -> abs (i-7) + abs(j-7) <= 3) b*)
    (*try choose_in (fun i j -> i <= 1 || i >= 14 || j <= 1 || j >= 14) b*)
    (*try choose_in (fun i j -> i <= 0 || i >= 15 || j <= 0 || j >= 15) b*)
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    try choose_in (fun i j -> i=0 || i=15 || j=0 || j=15) b
    with Not_found ->
    try choose_in (fun i j -> i=1 || i=14 || j=1 || j=14) b
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    with Not_found -> choose b

end

let dump b score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then 
  begin
  Board0.save b file;
  printf "saved in %s.@." file
  end

let solution b =
  if not greedy then begin
    printf "solution!@.";
    dump b 500
  end;
  ignore (read_key ());
  exit 0



let place ~complete b i j p =
  let b = Board0.place ~complete b i j p in
  if graphics then draw_tile i j p;
  let score = score b in
  if score > !best then begin
    printf "score = %d@." score;
    best := score;
    if score > 200 then dump b score
  end;
  b

let count = ref 0

(* look for two adjacent empty cells with incompatible possibilities *)
let check b =
  iter_empty
    (fun i j ->
       let pij = poss b i j in
       let test pr ps = Pset.for_all (fun p -> Pset.exists (pr p) ps) pij in
       if Pset.cardinal pij < 10 then begin
	 if i > 0 && piece_at b (i-1) j == dummy_piece then begin 
	   if not (test (fun p p' -> p.down == p'.up) (poss b (i-1) j)) then 
	     (printf "!@?"; raise Incomplete)
	 end;
	 if i < n-1 && piece_at b (i+1) j == dummy_piece then begin
	   if not (test (fun p p' -> p.up == p'.down) (poss b (i+1) j)) then
	     (printf "!@?"; raise Incomplete)
	 end;
	 if j > 0 && piece_at b i (j-1) == dummy_piece then begin
	   if not (test (fun p p' -> p.left == p'.right) (poss b i (j-1))) then
	     (printf "!@?"; raise Incomplete)
	 end;
	 if j < n-1 && piece_at b i (j+1) == dummy_piece then begin
	   if not (test (fun p p' -> p.right == p'.left) (poss b i (j+1))) then
	     (printf "!@?"; raise Incomplete)
	 end;
       end)
    b
*)

let () = Random.self_init ()

let rec loop b = ()
(*
  if todo b = 0 then solution b;
  (*incr count; if !count = 50 then begin count := 0; check b end;*)
  let i,j,ps = MinCard.choose b in
  Pset.random_iter
    (fun p -> 
       try 
	 let b = place ~complete b i j p in 
	 try loop b 
	 with Incorrect | Incomplete -> if graphics then clear_tile i j
       with Incorrect | Incomplete -> 
	 ())
    ps;
  if complete then raise Incomplete;
  (* only in greedy mode *)
  let b = place ~complete:false b i j dropped_piece in 
  loop b
*)

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let draw_board b =
  iter_grid 
    (fun i j -> 
       let p,r = get b i j in
       if p != dummy_tile then draw_tile i j p r else clear_tile i j)
  
let draw_rem x rem =
  let i = ref 15 in
  try
    TileSet.iter
      (fun p ->
	 draw_tile !i x p 0;
	 decr i;
	 if !i = 0 then raise Exit)
      rem;
    while !i > 0 do
      clear_tile !i x; decr i
    done
  with Exit -> ()
  
let init_board = Board0.load load

let rec fill_remaining b = 
  let b = ref b in
  let choose_put i j remaining =
    let (p,r) = get !b i j in
    if p.id < 0 then 
      let p = TileSet.choose remaining in
      b := put !b i j p 0
  in
  for i=1 to n-2 do
    for j=1 to n-2 do
      choose_put i j !b.remaining_others
    done
  done;
  for i=1 to n-2 do
    choose_put i 0 !b.remaining_edges;
    choose_put 0 i !b.remaining_edges;
    choose_put i (n-1) !b.remaining_edges;
    choose_put (n-1) i !b.remaining_edges;
  done;
  choose_put 0 0 !b.remaining_corners;
  choose_put 0 (n-1) !b.remaining_corners;
  choose_put (n-1) (n-1) !b.remaining_corners;
  choose_put (n-1) 0 !b.remaining_corners;
  !b

let check b = 
  let score = ref 0 in
  for i = 0 to n-1 do
    for j = 0 to n-2 do
      let (p,r) = get b i j in
      let (p',r') = get b i (j+1) in
      if is_real_tile p && is_real_tile p' && right_pat p r == left_pat p' r' 
      then incr score
      else vertical_line i j
    done;
    if i < n-1 then
      for j = 0 to n-1 do
	let (p,r) = get b i j in
	let (p',r') = get b (i+1) j in
	if is_real_tile p && is_real_tile p' && up_pat p r == down_pat p' r' 
	then incr score
	else hor_line i j
      done
  done;
  !score

let ij st = 
  let x = st.mouse_x in
  let y = st.mouse_y in
  if y >= step && y < step+n*step then
    (y - step) / step, (x - step) / step
  else
    -1,-1

let current_sel = ref None

let display b = 
  draw_board b;
  let s = check b in
  assert (s = b.score);
  draw_rem 19 b.remaining_corners;
  draw_rem 18 b.remaining_edges;
  draw_rem 17 b.remaining_others;
  set_window_title ("Editor3: score = " ^ (string_of_int b.score));
  printf "score = %d@." b.score;
  let f = (Filename.basename load) ^ string_of_int b.score
  in
  let file = Filename.concat sol_dir f in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    Board0.save b file;
    printf "saved in %s.@." file
  end;
  match !current_sel with
    | None -> ()
    | Some(i,j) -> border_tile i j
	

let fill b =
  let best = ref b.score in
  let best_b = ref b in
  let rec loop credit b =
    let pos = Array.create 5 [] in
    (* inside *)
    for i = 1 to n-2 do
      for j = 1 to n-2 do
	let p,_ = get b i j in
	if not (is_real_tile p) then 
	  TileSet.iter
	    (fun p ->
	       for r=0 to 3 do
		 let s = score_delta b i j p r in
		 pos.(s) <- (i,j,p,r) :: pos.(s)
	       done)
	    b.remaining_others
      done
    done;
    (* edges *)
    for i = 1 to n-2 do
      List.iter
	(fun (i,j,rot) -> 
	   let p,_ = get b i j in
	   if not (is_real_tile p) then 
	     TileSet.iter
	       (fun p ->
		  let r = match p.kind with
		    | Edge r -> (r+rot) mod 4
		    | _ -> assert false
		  in 
		  let s = score_delta b i j p r in
		  pos.(s) <- (i,j,p,r) :: pos.(s))
	       b.remaining_edges)
	[(0,i,0) ; (i,n-1,1) ; (n-1,i,2) ; (i,0,3)];
    done;
    (* corners *)
    List.iter
      (fun (i,j,rot) -> 
	 let p,_ = get b i j in
	 if not (is_real_tile p) then 
	   TileSet.iter
	     (fun p ->
		let r = match p.kind with
		  | Corner r -> (r+rot) mod 4
		  | _ -> assert false
		in 
		let s = score_delta b i j p r in
		pos.(s) <- (i,j,p,r) :: pos.(s))
	     b.remaining_corners)
      [(0,0,0) ; (0,n-1,1) ; (n-1,n-1,2) ; (n-1,0,3) ];
    let finished = ref true in
    let c = ref credit in
    try
      for s=4 downto 1 do
	finished := !finished && pos.(s) = [];
	List.iter
	  (fun (i,j,p,r) ->
	     let b = put b i j p r in 
	     loop !c b;
	     decr c; if !c <= 0 then raise Exit)
	  pos.(s)
      done;
      if !finished && b.score > ! best then 
	(display b; best := b.score ; best_b := b)
    with Exit -> ()
  in
  for credit = 1 to 5 do
    Format.printf "credit = %d@." credit;
    loop credit b;
  done;
  printf "filling done@.";
  !best_b
	    
      
let best_rot b i j p =
  match p.kind with
    | Corner r ->
	begin
	  match (i,j) with
	    | (0,0) -> r
	    | (0,_) -> (r+1) mod 4
	    | (_,0) -> (r+3) mod 4
	    | (_,_) -> (r+2) mod 4
	end
    | Edge r ->
	if i=0 then r 
	else if j=n-1 then (r+1) mod 4
	else if i=n-1 then (r+2) mod 4
	else if j=0 then (r+3) mod 4
	else assert false
    | Other ->
	let best_r = ref 0 in
	let best_d = ref 0 in
	for r = 0 to 3 do
	  let d = score_delta b i j p r in
	  if d > !best_d then (best_r := r; best_d := d)
	done;
	!best_r
	    
	
let () = 
  let b = ref init_board in
  printf "score = %d@." !b.score;
  display !b;
  while true do
    let st = wait_next_event [Button_down; Key_pressed; Mouse_motion] in
    if st.keypressed then begin match st.key with
      | 'q' -> exit 0
      | 'r' -> 
	  begin
	    match !current_sel with
	      | None -> ()
	      | Some(i,j) ->
		  let (p,r) = get !b i j in
		  b := remove !b i j;
		  b := put !b i j p ((r+1) mod 4);
		  display !b
	  end
      | 'e' ->
	  begin
	    match !current_sel with
	      | None -> ()
	      | Some(i,j) ->
		  let (i',j') = ij st in
		  if i <> i' || j <> j' then
		    let (p',r') = get !b i' j' in
		    let (p,r) = get !b i j in
		    if is_real_tile p then b := remove !b i j;
		    if is_real_tile p' then b := remove !b i' j';
		    if is_real_tile p' then 
		      begin
			let r' = best_rot !b i j p' in
			b := put !b i j p' r' ;
		      end;
		    if is_real_tile p then 
		      begin
			let r = best_rot !b i' j' p in
			b := put !b i' j' p r ;
		      end;
		    display !b
	  end
      | 'x' -> 	  
	  begin
	    match !current_sel with
	      | None -> ()
	      | Some(i,j) ->
		  let (p,r) = get !b i j in
		  if is_real_tile p then b := remove !b i j;
		  display !b
	  end
      | 'f' ->
	  b := fill !b;
	  display !b
      | 's' -> 
	  save !b "saved.sol"; printf "saved in saved.sol@."
      | _ -> ()
    end;
    if st.button then begin
      let i,j = ij st in
      current_sel := Some (i,j);
      display !b
    end
  done
