
open Format
open Brute7

let send_msg server m =
  Network.send server (Answer m)

let send_tree server t =
  Network.send server (Tree t)

let report_iteration server tile rot t =
  Brute7.report_iteration t;
  Network.send server (Report(tile,rot,score t,size t))

let interrupt_requested = ref false

let iteratively_expand_tree server first_tile first_rot b =
  let time = ref (Unix.time ()) in
  try
    while not !interrupt_requested do
      report_iteration server first_tile first_rot !b;
      b := expand_tree_root init_board first_tile first_rot !b;
      let t = Unix.time () in
      if t -. !time > 60.0 (* 1 minutes *) then
	begin
	  send_tree server (tree_to_ast !b);
	  time := t;
	end;
    done;
    report_iteration server first_tile first_rot !b;
    send_tree server (tree_to_ast !b);
    interrupt_requested := false;
    Network.send server Stopped
  with
      Exit ->
	assert false

let state = ref Idle

let handle_server_msg server msg =
  match msg with
    | DoCompute(t,x,y) ->
	begin
	  match !state with
	    | Idle -> 
		let t = ref t in
		let pid = 
		  Thread.create (iteratively_expand_tree server x y) t
		in
		state := Computing(pid,t,x,y)
	    | Computing(pid,t,x,y) -> 
		send_msg server "already computing"
	end
    | StopComputation ->
	begin
	  interrupt_requested := true;
	  send_msg server "got interrupt request, finishing iteration"	  
	end

	
let port = ref 2007

let () =
  let server = Network.connection ~machine:server_name ~port:!port in
  try
    while true do
      Network.receive (handle_server_msg server) server ~timeout:(-1.0)
    done
  with Network.Shutdown ->
    eprintf "server closed; bye@."
      
