
open Format
open Arg

let n = ref (-1)
let p = ref (-1)
let show = ref false
let seed = ref None

let spec = ["-size", Set_int n, "board width";
	    "-pats", Set_int p, "number of patterns";
	    "-show", Set show, "graphical display";
	    "-seed", Int (fun v -> seed := Some v), "random generator seed";]
let msg = "gen [options] -n <int> -p <int>"
let usage () = Arg.usage spec msg; exit 1
let () = parse spec (fun _ -> usage ()) msg

let n = !n
let p = !p
let () = if n = -1 || p = -1 then usage ()
let show = !show

let up = Array.create_matrix n n 0
let down = Array.create_matrix n n 0
let right = Array.create_matrix n n 0
let left = Array.create_matrix n n 0

(* 0    is the gray edges
   1..p are the different patterns *)

let () = match !seed with
  | None -> Random.self_init ()
  | Some v -> Random.init v

let () = 
  for i = 0 to n-1 do
    for j = 0 to n-2 do
      let k = 1 + Random.int p in
      right.(i).(j) <- k;
      left.(i).(j+1) <- k;
    done;
    if i < n-1 then
      for j = 0 to n-1 do
	let k = 1 + Random.int p in
	up.(i).(j) <- k;
	down.(i+1).(j) <- k;
      done
  done

let swap a i j = let t = a.(i) in a.(i) <- a.(j); a.(j) <- t

let () =
  printf "random@.";
  printf "%d@." n;
  let k = ref 0 in
  let a = Array.create (n*n) (0,0,0,0) in
  for i = 0 to n-1 do
    for j = 0 to n-1 do
      incr k;
      a.(!k-1) <- (right.(i).(j), up.(i).(j), left.(i).(j), down.(i).(j))
    done
  done;
  for k = 1 to n*n do
    let i = Random.int (n*n) in
    let j = Random.int (n*n) in
    swap a i j;
  done;
  for i = 0 to n*n-1 do
    let r,u,l,d = a.(i) in printf "%d %d %d %d %d@." (i+1) r u l d
  done

open Graphics
    
let step = 70
let half = step / 2

let color = function
  | 0 -> rgb 100 100 100
  | 1 -> red 
  | 2 -> green
  | 3 -> blue
  | 4 -> yellow
  | 5 -> cyan
  | 6 -> magenta
  | n -> assert (n <= 22); let k = 22-n in rgb (16 * k) 0 (255 - 16*k)

let () = if not show then exit 0

let () =
  open_graph (sprintf " %dx%d" (n * step) (n * step));
  for i = 0 to n-1 do
    for j = 0 to n-1 do
      let x = step * j + step/2 in
      let y = step * i + step/2 in
      Draw.down x y (step/2) down.(i).(j);
      Draw.right x y (step/2) right.(i).(j);
      Draw.up x y (step/2) up.(i).(j);
      Draw.left x y (step/2) left.(i).(j)
(***
      set_color (color down.(i).(j));
      fill_poly [| x,y; x+step,y; x+half,y+half |];
      set_color (color right.(i).(j));
      fill_poly [| x+step,y; x+half,y+half; x+step,y+step |];
      set_color (color up.(i).(j));
      fill_poly [| x+half,y+half; x+step,y+step; x,y+step |];
      set_color (color left.(i).(j));
      fill_poly [| x,y+step; x+half,y+half; x,y |];
***)
    done
  done;
  set_color black;
  for i = 1 to n-1 do
    moveto 0 (i*step); lineto (n*step) (i*step);
    moveto (i*step) 0; lineto (i*step) (n*step)
  done

let () = ignore (read_key ())

