
(**** specific implementation for 16x16 matrices 

type 'a t = 'a Parray.t

let init n m f =
  assert (n == 16 && m == 16);
  Parray.init 256 (fun k -> f (k lsr 4) (k land 0b1111))

let get m i j = 
  Parray.get m ((i lsl 4) lor j)

let set m i j v = 
  Parray.set m ((i lsl 4) lor j) v

let rec fold_n n f x =
  let rec loop acc i = if i = n then acc else loop (f acc i) (i+1) in
  loop x 0

let map f m0 = 
  fold_n 
    (Parray.length m0)
    (fun m k ->
       let v = f (k lsr 4) (k land 0b1111) (Parray.get m0 k) in
       Parray.set m k v)
    m0

****)

(* generic implementation *)

type 'a t = 'a Parray.t Parray.t

let init n m f =
  Parray.init n (fun i -> Parray.init m (f i))

let get m i j = 
  Parray.get (Parray.get m i) j

let set m i j v = 
  let rowi = Parray.get m i in
  Parray.set m i (Parray.set rowi j v)

let rec fold_n n f x =
  let rec loop acc i = if i = n then acc else loop (f acc i) (i+1) in
  loop x 0

let map f m0 =
  fold_n 
    (Parray.length m0)
    (fun m i ->
       let rowi0 = Parray.get m0 i in
       let rowi = 
	 fold_n 
	   (Parray.length rowi0)
	   (fun rowi j -> 
	      let vij = Parray.get rowi0 j in
	      Parray.set rowi j (f i j vij))
	   rowi0
       in
       Parray.set m i rowi)
    m0

