
(* Sets of pieces implemented as Patricia trees. *)

type t

type elt = Puzzle.piece

val empty : t

val is_empty : t -> bool

val mem : elt -> t -> bool

val add : elt -> t -> t

val singleton : elt -> t

val remove : elt -> t -> t

val union : t -> t -> t

val subset : t -> t -> bool

val inter : t -> t -> t

val diff : t -> t -> t

val equal : t -> t -> bool

val compare : t -> t -> int

val elements : t -> elt list

val choose : t -> elt

val cardinal : t -> int

val iter : (elt -> unit) -> t -> unit

val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a

val for_all : (elt -> bool) -> t -> bool

val exists : (elt -> bool) -> t -> bool

val filter : (elt -> bool) -> t -> t

val partition : (elt -> bool) -> t -> t * t

val split : elt -> t -> t * bool * t

val min_elt : t -> elt
val max_elt : t -> elt

val intersect : t -> t -> bool

(* additional functions for Eternity-II *)

val of_array : elt array -> t

val random_iter : (elt -> unit) -> t -> unit

val filter_opt : (elt -> bool) -> t -> t option (* None if unchanged *)

