
(* reads the puzzle from [file] if [-p file] specified on the command line,
   and from standard input otherwise *)

val n : int

val n2 : int (* n*n *)

val corner_distance : int -> int -> int

val is_eternity_ii : bool

type kind = Corner | Edge | Other

type piece = private
  { id : int; kind : kind; 
    right : int; up : int; left : int; down : int; (* patterns *)
    rot : int; (* 0,1,2,3 = ccw rotation wrt initial configuration *)
    uid : int; (* unique id used in Pset *)
  }

val rot : piece -> piece

val dummy_piece : piece (* empty slot *)

val dropped_piece : piece (* slot dropped in greedy search *)

val is_real_piece : piece -> bool

(* all [n*n] pieces, as read from standard input *)

val pieces : piece array
  (* [get_piece pid rot) returns the piece number pid rotated with [rot] *)
val get_piece : int -> int -> piece

(* save/load *)
val load : string -> piece array array
val save : string -> piece array array -> unit

(* the 4 corners *)

val corners_rot0 : piece array (* corner = top left *)
val corners_rot1 : piece array
val corners_rot2 : piece array
val corners_rot3 : piece array

(* the [4*n-8] edges *)

val edges_rot0 : piece array (* edge = top *)
val edges_rot1 : piece array 
val edges_rot2 : piece array 
val edges_rot3 : piece array 

(* the [(n-2)*(n-2)] other pieces *)

val others_rot0 : piece array
val others_rot1 : piece array
val others_rot2 : piece array
val others_rot3 : piece array

