
open Format
open Puzzle

(* adjacents *)

let adj = Hashtbl.create 97

let add p1 p2 =
  let p = p1,p2 in
  try incr (Hashtbl.find adj p) with Not_found -> Hashtbl.add adj p (ref 1)

(* face  face *)

let pairs = Hashtbl.create 97

let add_sym p1 p2 =
  let p = if p1 <= p2 then p1,p2 else p2,p1 in
  try incr (Hashtbl.find pairs p) with Not_found -> Hashtbl.add pairs p (ref 1)

let () = 
  Array.iter
    (fun p -> 
      add p.up p.left; add p.left p.down; add p.down p.right; add p.right p.up;
      add_sym p.up p.down; add_sym p.left p.right)
    pieces

let occ = Hashtbl.create 17

let () =
  printf "face a face@\n";
  printf "   "; for i = 1 to 22 do printf "%2d " i done; printf "@\n";
  printf "---------------------------------------------------------------------@\n";
  for i = 1 to 22 do
    printf "%2d:" i;
    for j = 1 to 22 do
      if j < i then printf "   " else 
	printf "%2d " (try !(Hashtbl.find pairs (i,j)) with Not_found -> 0)
    done;
    printf "@\n"
  done;
  printf "adjacents@\n";
  printf "   "; for i = 1 to 22 do printf "%2d " i done; printf "@\n";
  printf "---------------------------------------------------------------------@\n";
  for i = 1 to 22 do
    printf "%2d:" i;
    for j = 1 to 22 do
      let n = try !(Hashtbl.find adj (i,j)) with Not_found -> 0 in
      printf "%2d " n;
      begin 
	try incr (Hashtbl.find occ n) 
	with Not_found -> Hashtbl.add occ n (ref 1) 
      end;
    done;
    printf "@\n"
  done;
  printf "@."

let () =
  let l = Hashtbl.fold (fun n nn l -> (n, !nn) :: l) occ [] in
  let l = List.sort (fun (x,_) (y,_) -> x - y) l in
  List.iter (fun (n,nn) -> printf "%2d: %2d@." n nn) l
