
open Format
open Puzzle

let w = int_of_string Sys.argv.(1)
let h = int_of_string Sys.argv.(2)

let () = assert (n mod w = 0 && n mod h = 0)
let () = assert (w < n && h < n)

(* creating tiles of size wxh *)

type t_kind = 
    | UpLeft | UpRight | DownLeft | DownRight 
    | Right | Up | Left | Down
    | Middle

type tile = {
  t_kind : t_kind;
  t_right: int list; (* length h, from down to up *)
  t_up   : int list; (* length w, from right to left *)
  t_left : int list; (* length h, from down to up *)
  t_down : int list; (* length w, from right to left *)
  t_uses : piece list;
}

module S = Set.Make(struct type t = int let compare = Pervasives.compare end)

let count = ref 0

let current = Array.create_matrix h w dummy_piece

let all_tiles = Hashtbl.create 1000003

let new_tile () =
  let r = ref 0 in
  for i = 0 to h-1 do r := (!r lsl 5) lor current.(i).(w-1).right done;
  let u = ref 0 in
  for j = w-1 downto 0 do u := (!u lsl 5) lor current.(h-1).(j).up done;
  let l = ref 0 in
  for i = 0 to h-1 do l := (!l lsl 5) lor current.(i).(0).left done;
  let d = ref 0 in
  for j = w-1 downto 0 do d := (!d lsl 5) lor current.(0).(j).down done;
  let t = (!r,!u,!l,!d) in
  if not (Hashtbl.mem all_tiles t) then begin
    incr count;
    if !count mod 100000 = 0 then printf "%d @?" !count;
    Hashtbl.add all_tiles t ()
  end

let make_tiles (iter_poss : int -> int -> (piece -> unit) -> unit) =
  (* try to put piece p at (i,j); raises Exit if not possible *)
  let put i j p =
    if j > 0 && current.(i).(j-1).right != p.left then raise Exit;
    if i > 0 && current.(i-1).(j).up != p.down then raise Exit;
    current.(i).(j) <- p
  in
  let rec fill used i j =
    if i = h then
      (* new tile *)
      new_tile ()
    else
      (* try all possibilities for cell (i,j) *)
      iter_poss i j 
	(fun p -> 
	  try 
	    if S.mem p.id used then raise Exit;
	    put i j p; 
	    let used = S.add p.id used in
	    if j = w-1 then fill used (i+1) 0 else fill used i (j+1);
	    current.(i).(j) <- dummy_piece
	  with Exit -> 
	    ())
  in
  fill S.empty 0 0

(* middle tiles *)

let () = 
  make_tiles 
    (fun i j f -> 
      Array.iter f others_rot0; Array.iter f others_rot1;
      Array.iter f others_rot2; Array.iter f others_rot3)

let () = printf "%d middle tiles@." !count

(* instantiating DLX *)

